/*  Sunpos

 Takes an FT2 file and adds columns to include Sun RA and dec.
 Robin Corbet, UMBC. 2010-12-09

 Updated for Moon position by Paul Ray
 Code to add Sun positions 

 Argument is: 
   1) input FT2 file name

This needs:
- FITSIO library
- CSPICE library from JPL


*/

#include "fitsio.h"  /* required by every program that uses CFITSIO  */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "SpiceUsr.h"

const double DEG2RAD = M_PI/180.0;
const double RAD2DEG = 180.0/M_PI;

int main(int argc, char *argv[])
{
  fitsfile *fptr_sc;       
  int status, anynul, i, timebin;
  int n_points, n_fields;
  double *start_time;
  float *sun_ra, *sun_dec;
  float *moon_ra, *moon_dec;
  char buf[120], comment[100], jdstr[250];
  int time_colnum, sun_ra_colnum,sun_dec_colnum, moon_ra_colnum,moon_dec_colnum;
  double jdref = 2451910.5007429;  /* JD(TDT) */
  SpiceDouble et, moonlt, sunlt, sunrad, moonrad;
  SpiceDouble moonpt[3], sunpt[3];


  double ra, dec, jd_tt;

  char spacecraft_file[250];

  status = 0;

  if (argc != 2) {
    fprintf(stdout, "wrong number of arguments\nUsage: %s <FT2_name>.fits",argv[0]);
    return(-10);
  }

  strcpy(spacecraft_file, argv[1]);

  fits_open_data(&fptr_sc, spacecraft_file, READWRITE, &status);

  /* read spacecraft header to calculate array size */

  fits_read_key(fptr_sc, TINT, "NAXIS2",
		&n_points, comment, &status);

  printf("number of points = %d\n", n_points);

  fits_read_key(fptr_sc, TINT, "TFIELDS",
		&n_fields, comment, &status);

  printf("number of fields = %d\n", n_fields);

  /* First check whether there's already a MOON column */
  fits_get_colnum(fptr_sc, CASEINSEN, "RA_MOON",
		  &moon_ra_colnum, &status);

  if (status == 0){
    printf("RA_MOON column already exists. File will not be modified.\n");
    fits_close_file(fptr_sc, &status);            /* close the output file */
    return(0);
  } else {
    printf("No RA_MOON column yet. Adding RA_MOON/DEC_MOON columns.\n");
    status = 0;
  }

  moon_ra_colnum = n_fields + 1;
  moon_dec_colnum = moon_ra_colnum + 1;

  /* Load SPICE kernels */
  furnsh_c("fermikernel.tm");

  /* Set error actions */
  erract_c("SET",250,"RETURN");

  /* allocate needed array space */

  start_time = (double *) calloc (n_points, sizeof(double));

  sun_ra = (float *) calloc (n_points, sizeof(float));
  sun_dec = (float *) calloc (n_points, sizeof(float));
  moon_ra = (float *) calloc (n_points, sizeof(float));
  moon_dec = (float *) calloc (n_points, sizeof(float));

  /* Get the start time values */

  fits_get_colnum(fptr_sc, CASEINSEN, "START",
		  &time_colnum, &status);

  fits_read_col(fptr_sc, TDOUBLE, time_colnum, 1,
		1, n_points, NULL, start_time, 
		&anynul, &status);

  /* Loop over rows in FT2 file here */
  for (i = 0; i < n_points; i++){

    /* convert MET to Julian Day */
    jd_tt = start_time[i]/86400.0 + jdref;

    /* Compute ET (ephemeris time) */
    sprintf(jdstr,"JD %.6f TDT",jd_tt);
    str2et_c(jdstr,&et);

    /* Compute Sun RA and Dec the old way */
#if 0
    snradc(jd_tt, &ra, &dec);
    sun_ra[i] = ra;
    sun_dec[i] = dec;
#endif

    /* Compute Sun RA and Dec with SPICE and compare */
    spkpos_c("SUN",et,"J2000","LT","-1033053",sunpt,&sunlt);
    reclat_c(sunpt, &sunrad, &ra, &dec);
    if (failed_c()) {
      printf("Sun position failed at et %.6f\n",et);
      ra = 0.0;
      dec = 0.0;
      reset_c();
    }
#if 0
    printf("SUN DIFF %.6f, %.6f deg\n",RAD2DEG*ra-sun_ra[i],RAD2DEG*dec-sun_dec[i]);
#endif
    sun_ra[i] = RAD2DEG*ra;
    sun_dec[i] = RAD2DEG*dec;

    /* Compute Moon RA and Dec */
    /*    printf("%.6f\n",et); */
    spkpos_c("MOON",et,"J2000","LT","-1033053",moonpt,&moonlt);
    reclat_c(moonpt, &moonrad, &ra, &dec);
    if (failed_c()) {
      printf("Moon position failed at et %.6f\n",et);
      ra = 0.0;
      dec = 0.0;
      reset_c();
    }
    moon_ra[i] = RAD2DEG*ra;
    moon_dec[i] = RAD2DEG*dec;

/* Crude RA range check */
	if(moon_ra[i] < 0.0){
		moon_ra[i] = moon_ra[i] + 360.0;
	}

  }

  fits_insert_col(fptr_sc, moon_ra_colnum, "RA_MOON",
		  "E", &status);

  fits_insert_col(fptr_sc, moon_dec_colnum, "DEC_MOON",
		  "E", &status);


  fits_write_col(fptr_sc, TFLOAT, moon_ra_colnum, 1,
		 1, n_points, moon_ra, &status); 

  fits_write_col(fptr_sc, TFLOAT, moon_dec_colnum, 1,
		 1, n_points, moon_dec, &status); 

  fits_close_file(fptr_sc, &status);            /* close the output file */
  
  fits_report_error(stderr, status);  /* print out any error messages */
  return (status);
}



/* Hacked FORTRAN code! */

void snradc(double jd,double *ra, double *dec)
/*
C THIS ROUTINE COMPUTES APPROXIMATE POSITION OF THE SUN FOR SPECIFIED
C       JULIAN DATE.
C ALGORITHM TAKEN FROM 1994 NAUTICAL ALMANAC PAGE C24.
C       ACCURATE TO 0.01 DEGREES FROM 1950 TO 2050
C
C ALL VARIABLES ARE REAL*8
C
C INPUT:
C       JD IS JULIAN DATE

C
C OUTPUT:
C       RA IS RIGHT ASCENSION IN DEGREES (J2000?)
C       DEC IS DECLINATION IN DEGREES
C
*/
{
  double rad, l, g, elon, obl, snl;
  double n;

  /* printf("jd = %g\n", jd); */
  rad = RAD2DEG;
  n = jd - 2451545.0;

  /* l is mean longitude of sun, corrected for aberration */
  l = fmod(280.466 + 0.9856474*n, 360.0);
  /* g is mean anomaly */
  g = fmod(357.528 + 0.9856003*n, 360.0);
  /* elon is ecliptic longitude */
  elon = l + sin(g/rad)*1.915 + sin(2.0 * g/rad)*0.02;
  /* e is obliquity of ecliptic */
  obl = 23.440 - (4.0e-7)*n;
  snl = sin(elon/rad);
  *ra = rad * atan2(snl*cos(obl/rad),cos(elon/rad));
  if(*ra < 0.0e0){*ra = *ra + 360.0;}
  *dec = rad * asin(sin(obl/rad)*snl);

  return;
}
